@@ -1,35 +1,55 @@
 from __future__ import annotations
 
+import collections
 import functools
 import time
 
-from contextlib import suppress
+from enum import IntEnum
 from typing import TYPE_CHECKING
 
 from poetry.core.packages.dependency import Dependency
 
-from poetry.mixology.failure import SolveFailure
+from poetry.mixology.failure import SolveFailureError
 from poetry.mixology.incompatibility import Incompatibility
-from poetry.mixology.incompatibility_cause import ConflictCause
-from poetry.mixology.incompatibility_cause import NoVersionsCause
-from poetry.mixology.incompatibility_cause import PackageNotFoundCause
-from poetry.mixology.incompatibility_cause import RootCause
+from poetry.mixology.incompatibility_cause import ConflictCauseError
+from poetry.mixology.incompatibility_cause import NoVersionsCauseError
+from poetry.mixology.incompatibility_cause import RootCauseError
 from poetry.mixology.partial_solution import PartialSolution
 from poetry.mixology.result import SolverResult
 from poetry.mixology.set_relation import SetRelation
 from poetry.mixology.term import Term
-from poetry.packages import DependencyPackage
+from poetry.packages import PackageCollection
 
 
 if TYPE_CHECKING:
     from poetry.core.packages.project_package import ProjectPackage
 
+    from poetry.packages import DependencyPackage
     from poetry.puzzle.provider import Provider
 
 
 _conflict = object()
 
 
+class Preference(IntEnum):
+    """
+    Preference is one of the criteria for choosing which dependency to solve
+    first. A higher value means that there are "more options" to satisfy
+    a dependency. A lower value takes precedence.
+    """
+
+    DIRECT_ORIGIN = 0
+    NO_CHOICE = 1
+    USE_LATEST = 2
+    LOCKED = 3
+    DEFAULT = 4
+
+
+CompKey = tuple[Preference, int, bool, int]
+
+DependencyCacheKey = tuple[str, str | None, str | None, str | None, str | None]
+
+
 class DependencyCache:
     """
     A cache of the valid dependencies.
@@ -40,37 +60,88 @@ class DependencyCache:
     """
 
     def __init__(self, provider: Provider) -> None:
-        self.provider = provider
-        self.cache: dict[
-            tuple[str, str | None, str | None, str | None, str | None],
-            list[DependencyPackage],
-        ] = {}
+        self._provider = provider
 
-        self.search_for = functools.lru_cache(maxsize=128)(self._search_for)
+        # self._cache maps a package name to a stack of cached package lists,
+        # ordered by the decision level which added them to the cache. This is
+        # done so that when backtracking we can maintain cache entries from
+        # previous decision levels, while clearing cache entries from only the
+        # rolled back levels.
+        #
+        # In order to maintain the integrity of the cache, `clear_level()`
+        # needs to be called in descending order as decision levels are
+        # backtracked so that the correct items can be popped from the stack.
+        self._cache: dict[DependencyCacheKey, list[list[DependencyPackage]]] = (
+            collections.defaultdict(list)
+        )
+        self._cached_dependencies_by_level: dict[int, list[DependencyCacheKey]] = (
+            collections.defaultdict(list)
+        )
 
-    def _search_for(self, dependency: Dependency) -> list[DependencyPackage]:
+        self._search_for_cached = functools.lru_cache(maxsize=128)(self._search_for)
+
+    def _search_for(
+        self,
+        dependency: Dependency,
+        key: DependencyCacheKey,
+    ) -> list[DependencyPackage]:
+        cache_entries = self._cache[key]
+        if cache_entries:
+            packages = [
+                p
+                for p in cache_entries[-1]
+                if dependency.constraint.allows(p.package.version)
+            ]
+        else:
+            packages = None
+
+        # provider.search_for() normally does not include pre-release packages
+        # (unless requested), but will include them if there are no other
+        # eligible package versions for a version constraint.
+        #
+        # Therefore, if the eligible versions have been filtered down to
+        # nothing, we need to call provider.search_for() again as it may return
+        # additional results this time.
+        if not packages:
+            packages = self._provider.search_for(dependency)
+
+        return packages
+
+    def search_for(
+        self,
+        dependency: Dependency,
+        decision_level: int,
+    ) -> list[DependencyPackage]:
         key = (
-            dependency.complete_name,
+            dependency.name,
             dependency.source_type,
             dependency.source_url,
             dependency.source_reference,
             dependency.source_subdirectory,
         )
 
-        packages = self.cache.get(key)
-        if packages is None:
-            packages = self.provider.search_for(dependency)
-        else:
-            packages = [
-                p for p in packages if dependency.constraint.allows(p.package.version)
-            ]
+        # We could always use dependency.without_features() here,
+        # but for performance reasons we only do it if necessary.
+        packages = self._search_for_cached(
+            dependency.without_features() if dependency.features else dependency, key
+        )
+        if not self._cache[key] or self._cache[key][-1] is not packages:
+            self._cache[key].append(packages)
+            self._cached_dependencies_by_level[decision_level].append(key)
 
-        self.cache[key] = packages
+        if dependency.features and packages:
+            # Use the cached dependency so that a possible explicit source is set.
+            return PackageCollection(
+                packages[0].dependency.with_features(dependency.features), packages
+            )
 
         return packages
 
-    def clear(self) -> None:
-        self.cache.clear()
+    def clear_level(self, level: int) -> None:
+        if level in self._cached_dependencies_by_level:
+            self._search_for_cached.cache_clear()
+            for key in self._cached_dependencies_by_level.pop(level):
+                self._cache[key].pop()
 
 
 class VersionSolver:
@@ -82,26 +153,17 @@ class VersionSolver:
     on how this solver works.
     """
 
-    def __init__(
-        self,
-        root: ProjectPackage,
-        provider: Provider,
-        locked: dict[str, list[DependencyPackage]] | None = None,
-        use_latest: list[str] | None = None,
-    ) -> None:
+    def __init__(self, root: ProjectPackage, provider: Provider) -> None:
         self._root = root
         self._provider = provider
         self._dependency_cache = DependencyCache(provider)
-        self._locked = locked or {}
-
-        if use_latest is None:
-            use_latest = []
-
-        self._use_latest = use_latest
-
         self._incompatibilities: dict[str, list[Incompatibility]] = {}
         self._contradicted_incompatibilities: set[Incompatibility] = set()
+        self._contradicted_incompatibilities_by_level: dict[
+            int, set[Incompatibility]
+        ] = collections.defaultdict(set)
         self._solution = PartialSolution()
+        self._get_comp_key_cached = functools.cache(self._get_comp_key)
 
     @property
     def solution(self) -> PartialSolution:
@@ -117,7 +179,7 @@ class VersionSolver:
         root_dependency.is_root = True
 
         self._add_incompatibility(
-            Incompatibility([Term(root_dependency, False)], RootCause())
+            Incompatibility([Term(root_dependency, False)], RootCauseError())
         )
 
         try:
@@ -168,10 +230,16 @@ class VersionSolver:
                     # decision level, so we clear [changed] and refill it with the
                     # newly-propagated assignment.
                     changed.clear()
-                    changed.add(str(self._propagate_incompatibility(root_cause)))
+                    result = self._propagate_incompatibility(root_cause)
+                    assert result is not None
+                    assert result != _conflict
+                    assert isinstance(result, str)
+                    changed.add(result)
                     break
-                elif result is not None:
-                    changed.add(str(result))
+
+                if result is not None:
+                    assert isinstance(result, str)
+                    changed.add(result)
 
     def _propagate_incompatibility(
         self, incompatibility: Incompatibility
@@ -199,6 +267,9 @@ class VersionSolver:
                 # incompatibility is contradicted as well and there's nothing new we
                 # can deduce from it.
                 self._contradicted_incompatibilities.add(incompatibility)
+                self._contradicted_incompatibilities_by_level[
+                    self._solution.decision_level
+                ].add(incompatibility)
                 return None
             elif relation == SetRelation.OVERLAPPING:
                 # If more than one term is inconclusive, we can't deduce anything about
@@ -217,6 +288,9 @@ class VersionSolver:
             return _conflict
 
         self._contradicted_incompatibilities.add(incompatibility)
+        self._contradicted_incompatibilities_by_level[
+            self._solution.decision_level
+        ].add(incompatibility)
 
         adverb = "not " if unsatisfied.is_positive() else ""
         self._log(f"derived: {adverb}{unsatisfied.dependency}")
@@ -310,9 +384,16 @@ class VersionSolver:
                 previous_satisfier_level < most_recent_satisfier.decision_level
                 or most_recent_satisfier.cause is None
             ):
+                for level in range(
+                    self._solution.decision_level, previous_satisfier_level, -1
+                ):
+                    if level in self._contradicted_incompatibilities_by_level:
+                        self._contradicted_incompatibilities.difference_update(
+                            self._contradicted_incompatibilities_by_level.pop(level),
+                        )
+                    self._dependency_cache.clear_level(level)
+
                 self._solution.backtrack(previous_satisfier_level)
-                self._contradicted_incompatibilities.clear()
-                self._dependency_cache.clear()
                 if new_incompatibility:
                     self._add_incompatibility(incompatibility)
 
@@ -335,8 +416,8 @@ class VersionSolver:
             # The most_recent_satisfier may not satisfy most_recent_term on its own
             # if there are a collection of constraints on most_recent_term that
             # only satisfy it together. For example, if most_recent_term is
-            # `foo ^1.0.0` and _solution contains `[foo >=1.0.0,
-            # foo <2.0.0]`, then most_recent_satisfier will be `foo <2.0.0` even
+            # `foo ^1.0.0` and _solution contains `[foo >=1.0.0,
+            # foo <2.0.0]`, then most_recent_satisfier will be `foo <2.0.0` even
             # though it doesn't totally satisfy `foo ^1.0.0`.
             #
             # In this case, we add `not (most_recent_satisfier \ most_recent_term)` to
@@ -344,12 +425,15 @@ class VersionSolver:
             # details.
             #
             # .. _algorithm documentation:
-            # https://github.com/dart-lang/pub/tree/master/doc/solver.md#conflict-resolution  # noqa: E501
+            # https://github.com/dart-lang/pub/tree/master/doc/solver.md#conflict-resolution
             if difference is not None:
-                new_terms.append(difference.inverse)
+                inverse = difference.inverse
+                if inverse.dependency != most_recent_satisfier.dependency:
+                    new_terms.append(inverse)
 
             incompatibility = Incompatibility(
-                new_terms, ConflictCause(incompatibility, most_recent_satisfier.cause)
+                new_terms,
+                ConflictCauseError(incompatibility, most_recent_satisfier.cause),
             )
             new_incompatibility = True
 
@@ -361,7 +445,118 @@ class VersionSolver:
             self._log(f'! which is caused by "{most_recent_satisfier.cause}"')
             self._log(f"! thus: {incompatibility}")
 
-        raise SolveFailure(incompatibility)
+        raise SolveFailureError(incompatibility)
+
+    def _get_comp_key(self, dependency: Dependency) -> CompKey:
+        """
+        Returns a tuple of
+        - preference
+        - num_deps_upper_bound
+        - has_deps
+        - num_packages
+        that serves as priority for choosing the next package to resolve.
+        (A lower value takes precedence.)
+
+        In order to provide results that are as deterministic as possible
+        and consistent between `poetry lock` and `poetry update`, the return value
+        of two different dependencies should not be equal if possible.
+
+        ## preference
+
+        See Preference class.
+
+        ## num_deps_upper_bound
+
+        A dependency with an upper bound is more likely to cause conflicts. Therefore,
+        a package with more dependencies with upper bounds should be chosen first.
+
+        ## has_deps
+
+        A package with dependencies should be chosen first
+        because a package without dependencies is less likely to cause conflicts.
+
+        ## num_packages
+
+        The original algorithm proposes to prefer packages with as few remaining
+        versions as possible, so that if a conflict is necessary it's forced quickly.
+        https://github.com/dart-lang/pub/blob/master/doc/solver.md#decision-making
+        However, this leads to the famous boto3 vs. urllib3 issue, so we prefer
+        packages with more remaining versions (see
+        https://github.com/python-poetry/poetry/pull/8255#issuecomment-1657198242
+        for more details).
+        """
+        preference = Preference.DEFAULT
+
+        # Direct origin dependencies must be handled first: we don't want to resolve
+        # a regular dependency for some package only to find later that we had a
+        # direct-origin dependency.
+        if dependency.is_direct_origin():
+            preference = Preference.DIRECT_ORIGIN
+
+        packages: list[DependencyPackage] = []
+        use_latest = dependency.name in self._provider.use_latest
+        if not use_latest:
+            locked = self._provider.get_locked(dependency)
+            if locked:
+                if preference == Preference.DEFAULT:
+                    preference = Preference.LOCKED
+                packages = [locked]
+
+        if not packages:
+            packages = self._dependency_cache.search_for(
+                dependency, self._solution.decision_level
+            )
+        num_packages = len(packages)
+        if packages:
+            package = packages[0].package
+            if package.is_root():
+                relevant_dependencies = package.all_requires
+            else:
+                if preference != Preference.LOCKED and not package.is_direct_origin():
+                    # We have to get the package from the pool,
+                    # otherwise `requires` will be empty.
+                    #
+                    # We might need `package.source_reference` as fallback
+                    # for transitive dependencies without a source
+                    # if there is a top-level dependency
+                    # for the same package with an explicit source.
+                    for repo in (dependency.source_name, package.source_reference):
+                        try:
+                            package = self._provider.get_package_from_pool(
+                                package.pretty_name,
+                                package.version,
+                                repository_name=repo,
+                            )
+                        except Exception:
+                            pass
+                        else:
+                            break
+
+                relevant_dependencies = [
+                    r
+                    for r in package.requires
+                    if not r.in_extras or r.in_extras[0] in dependency.extras
+                ]
+            has_deps = bool(relevant_dependencies)
+            num_deps_upper_bound = sum(
+                1 for d in relevant_dependencies if d.constraint.has_upper_bound()
+            )
+        else:
+            has_deps = False
+            num_deps_upper_bound = 0
+
+        if preference == Preference.DEFAULT:
+            if num_packages < 2:
+                preference = Preference.NO_CHOICE
+            elif use_latest:
+                preference = Preference.USE_LATEST
+        return preference, -num_deps_upper_bound, not has_deps, -num_packages
+
+    def _choose_next(self, unsatisfied: list[Dependency]) -> Dependency:
+        """
+        Chooses the next package to resolve.
+        """
+        return min(unsatisfied, key=self._get_comp_key_cached)
 
     def _choose_package_version(self) -> str | None:
         """
@@ -375,71 +570,26 @@ class VersionSolver:
         if not unsatisfied:
             return None
 
-        # Prefer packages with as few remaining versions as possible,
-        # so that if a conflict is necessary it's forced quickly.
-        def _get_min(dependency: Dependency) -> tuple[bool, int]:
-            # Direct origin dependencies must be handled first: we don't want to resolve
-            # a regular dependency for some package only to find later that we had a
-            # direct-origin dependency.
-            if dependency.is_direct_origin():
-                return False, -1
+        dependency = self._choose_next(unsatisfied)
 
-            if dependency.name in self._use_latest:
-                # If we're forced to use the latest version of a package, it effectively
-                # only has one version to choose from.
-                return not dependency.marker.is_any(), 1
-
-            locked = self._get_locked(dependency)
-            if locked:
-                return not dependency.marker.is_any(), 1
-
-            try:
-                return (
-                    not dependency.marker.is_any(),
-                    len(self._dependency_cache.search_for(dependency)),
-                )
-            except ValueError:
-                return not dependency.marker.is_any(), 0
-
-        if len(unsatisfied) == 1:
-            dependency = unsatisfied[0]
-        else:
-            dependency = min(*unsatisfied, key=_get_min)
-
-        locked = self._get_locked(dependency)
+        locked = self._provider.get_locked(dependency)
         if locked is None:
-            try:
-                packages = self._dependency_cache.search_for(dependency)
-            except ValueError as e:
-                self._add_incompatibility(
-                    Incompatibility([Term(dependency, True)], PackageNotFoundCause(e))
-                )
-                complete_name: str = dependency.complete_name
-                return complete_name
-
-            package = None
-            if locked is not None:
-                package = next(
-                    (
-                        p
-                        for p in packages
-                        if p.package.version == locked.package.version
-                    ),
-                    None,
-                )
-            if package is None:
-                with suppress(IndexError):
-                    package = packages[0]
+            packages = self._dependency_cache.search_for(
+                dependency, self._solution.decision_level
+            )
+            package = next(iter(packages), None)
 
             if package is None:
                 # If there are no versions that satisfy the constraint,
                 # add an incompatibility that indicates that.
                 self._add_incompatibility(
-                    Incompatibility([Term(dependency, True)], NoVersionsCause())
+                    Incompatibility([Term(dependency, True)], NoVersionsCauseError())
                 )
 
                 complete_name = dependency.complete_name
                 return complete_name
+
+            package.dependency.transitive_marker = dependency.transitive_marker
         else:
             package = locked
 
@@ -499,26 +649,5 @@ class VersionSolver:
                 incompatibility
             )
 
-    def _get_locked(self, dependency: Dependency) -> DependencyPackage | None:
-        if dependency.name in self._use_latest:
-            return None
-
-        locked = self._locked.get(dependency.name, [])
-        for dependency_package in locked:
-            package = dependency_package.package
-            if (
-                # Locked dependencies are always without features.
-                # Thus, we can't use is_same_package_as() here because it compares
-                # the complete_name (including features).
-                dependency.name == package.name
-                and (
-                    dependency.source_type is None
-                    or dependency.is_same_source_as(package)
-                )
-                and dependency.constraint.allows(package.version)
-            ):
-                return DependencyPackage(dependency, package)
-        return None
-
     def _log(self, text: str) -> None:
         self._provider.debug(text, self._solution.attempted_solutions)
