@@ -6,6 +6,7 @@ import hashlib
 from contextlib import contextmanager
 from contextlib import suppress
 from pathlib import Path
+from tempfile import TemporaryDirectory
 from typing import TYPE_CHECKING
 from typing import Any
 
@@ -15,7 +16,6 @@ import requests.adapters
 from packaging.metadata import parse_email
 from poetry.core.constraints.version import parse_constraint
 from poetry.core.packages.dependency import Dependency
-from poetry.core.utils.helpers import temporary_directory
 from poetry.core.version.markers import parse_marker
 
 from poetry.config.config import Config
@@ -26,6 +26,7 @@ from poetry.repositories.cached_repository import CachedRepository
 from poetry.repositories.exceptions import PackageNotFoundError
 from poetry.repositories.exceptions import RepositoryError
 from poetry.repositories.link_sources.html import HTMLPage
+from poetry.repositories.link_sources.json import SimpleJsonPage
 from poetry.utils.authenticator import Authenticator
 from poetry.utils.constants import REQUESTS_TIMEOUT
 from poetry.utils.helpers import HTTPRangeRequestSupportedError
@@ -38,6 +39,7 @@ if TYPE_CHECKING:
     from collections.abc import Iterator
 
     from packaging.utils import NormalizedName
+    from poetry.core.packages.package import PackageFile
     from poetry.core.packages.utils.link import Link
 
     from poetry.repositories.link_sources.base import LinkSource
@@ -110,7 +112,7 @@ class HTTPRepository(CachedRepository):
         self, link: Link, *, raise_accepts_ranges: bool = False
     ) -> Iterator[Path]:
         self._log(f"Downloading: {link.url}", level="debug")
-        with temporary_directory() as temp_dir:
+        with TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
             filepath = Path(temp_dir) / link.filename
             self._download(
                 link.url, filepath, raise_accepts_ranges=raise_accepts_ranges
@@ -340,7 +342,7 @@ class HTTPRepository(CachedRepository):
                 f' "{data.version}"'
             )
 
-        files: list[dict[str, Any]] = []
+        files: list[PackageFile] = []
         for link in links:
             if link.yanked and not data.yanked:
                 # drop yanked files unless the entire release is yanked
@@ -359,7 +361,33 @@ class HTTPRepository(CachedRepository):
             ):
                 file_hash = f"{hash_type}:{link.hashes[hash_type]}"
 
-            files.append({"file": link.filename, "hash": file_hash})
+            if file_hash is None:
+                # Is that even possible?
+                # Before introducing this warning and ignoring the file,
+                # null hashes would have been written to the lockfile,
+                # which should have been failed in the Chooser at latest.
+                self._log(
+                    f"Failed to determine hash of {link.url}. Skipping file.",
+                    level="warning",
+                )
+            else:
+                files.append(
+                    {
+                        "file": link.filename,
+                        "hash": file_hash,
+                        "url": link.url_without_fragment,
+                    }
+                )
+                if link.size is not None:
+                    files[-1]["size"] = link.size
+                if link.upload_time_isoformat is not None:
+                    files[-1]["upload_time"] = link.upload_time_isoformat
+
+        if not files:
+            raise PackageNotFoundError(
+                f'Could not determine a hash for any distribution link of package: "{data.name}" version:'
+                f' "{data.version}"'
+            )
 
         data.files = files
 
@@ -400,11 +428,13 @@ class HTTPRepository(CachedRepository):
                 return f"{required_hash.name}:{required_hash.hexdigest()}"
         return None
 
-    def _get_response(self, endpoint: str) -> requests.Response | None:
+    def _get_response(
+        self, endpoint: str, *, headers: dict[str, str] | None = None
+    ) -> requests.Response | None:
         url = self._url + endpoint
         try:
             response: requests.Response = self.session.get(
-                url, raise_for_status=False, timeout=REQUESTS_TIMEOUT
+                url, raise_for_status=False, timeout=REQUESTS_TIMEOUT, headers=headers
             )
             if response.status_code in (401, 403):
                 self._log(
@@ -425,8 +455,25 @@ class HTTPRepository(CachedRepository):
             )
         return response
 
+    def _get_prefer_json_header(self) -> dict[str, str]:
+        # Prefer json, but accept anything for backwards compatibility.
+        # Although the more specific value should be preferred to the less specific one
+        # according to https://developer.mozilla.org/en-US/docs/Glossary/Quality_values,
+        # we add a quality value because some servers still prefer html without one.
+        return {"Accept": "application/vnd.pypi.simple.v1+json, */*;q=0.1"}
+
+    def _is_json_response(self, response: requests.Response) -> bool:
+        return (
+            response.headers.get("Content-Type", "").split(";")[0].strip()
+            == "application/vnd.pypi.simple.v1+json"
+        )
+
     def _get_page(self, name: NormalizedName) -> LinkSource:
-        response = self._get_response(f"/{name}/")
+        response = self._get_response(
+            f"/{name}/", headers=self._get_prefer_json_header()
+        )
         if not response:
             raise PackageNotFoundError(f"Package [{name}] not found.")
+        if self._is_json_response(response):
+            return SimpleJsonPage(response.url, response.json())
         return HTMLPage(response.url, response.text)
